#include "screen.h"
#include "ports.h"
#include "../kernel/util.h"

// Offset depends on the type of the video_memory pointer, if it is short: 1screenChar -> 1offset

void clear_screen() {
    int screen_size = MAX_COLS * MAX_ROWS - 1;
    int i;
    short* video_memory = (short*) VIDEO_ADDRESS;

    // Set each video memory address to '' value with WHITE_ON_BLACK attribute
    short character_and_attribute_byte = 0 | WHITE_ON_BLACK << 8;
    for (i = 0; i < screen_size; i++) {
        video_memory[i] = character_and_attribute_byte;
    }
    
    // Set cursor to first offset after clear screen
    set_cursor_offset(0);
}

// Convert input coordinates to video memory offset
int get_offset(int row, int col) {
    return (row * MAX_COLS + col);
}

// Set cursor position to specified video_memory
void set_cursor_offset(int offset) {
    port_byte_out(REG_SCREEN_CTRL, 14);
    port_byte_out(REG_SCREEN_DATA, offset >> 8); // High byte
    port_byte_out(REG_SCREEN_CTRL, 15);
    port_byte_out(REG_SCREEN_DATA, (offset & 0xff)); // Low byte
}

// Get cursor position and return video_memory offset equivalent
int get_cursor_offset() {
    port_byte_out(REG_SCREEN_CTRL, 14);
    int offset = port_byte_in(REG_SCREEN_DATA) << 8; // High byte
    port_byte_out(REG_SCREEN_CTRL, 15); // Low byte
    offset += port_byte_in(REG_SCREEN_DATA);
    return offset; /* Position * size of character cell */
}

int print_char(char character, int row, int col, char attribute_byte) {
    short* video_memory = (short*) VIDEO_ADDRESS;
    int offset;

    if (!attribute_byte) {
        attribute_byte = WHITE_ON_BLACK;
    }
    
    //short code variable improves printing chars performance, in that way we only write to video_memory one time, not twice with char+attr
    short character_and_attribute_byte = character | attribute_byte << 8;
    
    if (row >= 0 && col >= 0) {
        offset = get_offset(row, col);
    } else {
        offset = get_cursor_offset();
    }


    // If we see a newline character, set offset to the end of current row, so it will be advanced to the first col
    // of the next row when offet is incremented.
    if (character == '\n') {
        int row = get_offset_row(offset);
        offset = get_offset(row, MAX_COLS - 1);
    } else {
        //video_memory[offset] = character;
        //video_memory[offset+1] = attribute_byte;
        video_memory[offset] = character_and_attribute_byte;
    }

    // Update the offset to the next character cell
    offset++;
    
    // Make scrolling adjustment, we check if next char write will be done outside screen char matrix
    offset = handle_scrolling(offset);
    
    // Update the cursor position on the screen device .
    set_cursor_offset(offset);
    
    return offset;
}

void print_string(char *message, int row, int col) {
    int offset;
    
    /* Set cursor if col/row are negative */
    if (row < 0 || col < 0 ) {
        offset = get_cursor_offset();
        row = get_offset_row(offset);
        col = get_offset_col(offset, row);
    }
    
    /* Loop through message and print it */
    int i = 0;
    while (message[i] != 0) {
        offset = print_char(message[i], row, col, WHITE_ON_BLACK);
        i++;
        /* Compute row/col for next iteration */
        row = get_offset_row(offset);
        col = get_offset_col(offset, row);
    }
}

int get_offset_row(int offset) {
    return offset / MAX_COLS;
}

int get_offset_col(int offset, int row) {
    return (offset - (row * MAX_COLS));
}

/* Advance the text cursor, scrolling the video buffer if necessary. */
int handle_scrolling(int offset) {
    // Are we writting inside screen char matrix?
    if (offset < MAX_ROWS * MAX_COLS) {
        return offset ;
    }
    
    // We are outside screen char matrix so copy row to inmediately above.
    int i;
    int row_length = (MAX_COLS-1) * 2;
    // While we are working with memory addresses bear in mind that offset = 1short(2bytes), memory_pos = offset * 2
    for (i = 1; i < MAX_ROWS - 1; i++) {
        memory_copy((short*) (VIDEO_ADDRESS + (get_offset(i, 0) * 2)), (short*) (VIDEO_ADDRESS + (get_offset(i-1, 0) * 2)), row_length);
    }
    
    // Blank last line by setting all bytes to 0
    short* last_line = (short*) (VIDEO_ADDRESS + (get_offset(MAX_ROWS-1, 0) * 2));
    short character_and_attribute_byte = 0 | WHITE_ON_BLACK << 8;
    for (i = 0; i < (MAX_COLS - 1); i++) {
        last_line[i] = character_and_attribute_byte;
    }
    
    // Offset outside screen char matrix, move the offset back one row.
    offset -= MAX_COLS ;
    
    // Return the updated cursor position .
    return offset ;
}
